searchData={"items":[{"type":"module","title":"SagaWeaver","doc":"The `SagaWeaver` module serves as the primary interface for executing and retrieving sagas within the SagaWeaver framework. It provides functions to orchestrate sagas based on incoming messages and to access existing saga instances.","ref":"SagaWeaver.html"},{"type":"module","title":"Overview - SagaWeaver","doc":"- **execute_saga/2**: Executes a saga by processing a given message. It determines whether to start a new saga or continue an existing one based on the message.\n- **retrieve_saga/2**: Retrieves an existing saga instance associated with a message.","ref":"SagaWeaver.html#module-overview"},{"type":"module","title":"Usage - SagaWeaver","doc":"To use `SagaWeaver`, you typically have saga modules that implement the saga logic using `use SagaWeaver.Saga`. You can then execute or retrieve sagas using the functions provided by this module.","ref":"SagaWeaver.html#module-usage"},{"type":"module","title":"Executing a Saga - SagaWeaver","doc":"The `execute_saga/2` function is used to process a message within the context of a saga. It will either start a new saga or continue an existing one.\n\n**Example:**\n\n```elixir\nalias MyApp.Sagas.OrderSaga\nalias MyApp.Events.OrderPlaced\n\nmessage = %OrderPlaced{order_id: 123, customer_id: 456}\n\ncase SagaWeaver.execute_saga(OrderSaga, message) do\n  {:ok, saga_instance} ->\n    # Saga executed successfully\n    IO.inspect(saga_instance, label: \"Saga Instance\")\n\n  {:noop, reason} ->\n    # No operation was performed\nend\n```","ref":"SagaWeaver.html#module-executing-a-saga"},{"type":"module","title":"Retrieving a Saga - SagaWeaver","doc":"The retrieve_saga/2 function allows you to fetch an existing saga instance based on a message. This is useful if you need to access the saga's state or context outside of the execution flow.\n\nExample:\n```elixir\nalias MyApp.Sagas.OrderSaga\nalias MyApp.Events.OrderUpdated\n\nmessage = %OrderUpdated{order_id: 123}\n\ncase SagaWeaver.retrieve_saga(OrderSaga, message) do\n{:ok, saga_instance} ->\n  # Saga instance retrieved\n  IO.inspect(saga_instance, label: \"Retrieved Saga Instance\")\n\n{:ok, :not_found} ->\n  # No saga found for the given message\n  IO.puts(\"Saga not found\")\n\n{:error, reason} ->\n  # An error occurred\nend`\n\n```","ref":"SagaWeaver.html#module-retrieving-a-saga"},{"type":"module","title":"Notes - SagaWeaver","doc":"- Ensure that your saga modules are properly defined using use SagaWeaver.Saga and implement the necessary callbacks.\n- The execute_saga/2 function internally uses the SagaWeaver.Orchestrator to manage saga execution.\n- The uniqueness of a saga instance is determined by the message and the saga's identity mapping. Make sure your identity mappings are correctly configured.","ref":"SagaWeaver.html#module-notes"},{"type":"module","title":"Functions - SagaWeaver","doc":"- execute_saga/2: Executes or continues a saga based on the provided message.\n- retrieve_saga/2: Retrieves an existing saga instance associated with a messag","ref":"SagaWeaver.html#module-functions"},{"type":"function","title":"SagaWeaver.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"SagaWeaver.html#child_spec/1"},{"type":"function","title":"SagaWeaver.execute_saga/2","doc":"","ref":"SagaWeaver.html#execute_saga/2"},{"type":"function","title":"SagaWeaver.retrieve_saga/2","doc":"","ref":"SagaWeaver.html#retrieve_saga/2"},{"type":"function","title":"SagaWeaver.start_link/1","doc":"","ref":"SagaWeaver.html#start_link/1"},{"type":"behaviour","title":"SagaWeaver.Adapters.StorageAdapter","doc":"The `SagaWeaver.Adapters.StorageAdapter` module defines the behaviour and provides an interface for storage adapters used by the SagaWeaver framework.","ref":"SagaWeaver.Adapters.StorageAdapter.html"},{"type":"behaviour","title":"Overview - SagaWeaver.Adapters.StorageAdapter","doc":"SagaWeaver allows for pluggable storage backends to persist saga state and context. This module specifies the required callbacks that any storage adapter must implement to integrate with the framework. By abstracting the storage layer, SagaWeaver can support different storage systems like Redis, PostgreSQL, or custom solutions.","ref":"SagaWeaver.Adapters.StorageAdapter.html#module-overview"},{"type":"behaviour","title":"Responsibilities - SagaWeaver.Adapters.StorageAdapter","doc":"- **Define Behaviour**: Specifies the set of functions (callbacks) that a storage adapter must implement.\n- **Delegation**: Provides default implementations that delegate function calls to the configured storage adapter module.","ref":"SagaWeaver.Adapters.StorageAdapter.html#module-responsibilities"},{"type":"behaviour","title":"Callbacks - SagaWeaver.Adapters.StorageAdapter","doc":"Storage adapters must implement the following callbacks:\n\n- `initialize_saga/1`: Initializes and persists a new saga instance.\n- `saga_exists?/1`: Checks if a saga with the given identifier exists.\n- `get_saga/1`: Retrieves a saga instance by its unique identifier.\n- `mark_as_completed/1`: Marks a saga as completed without deleting it.\n- `complete_saga/1`: Completes and removes a saga from storage.\n- `assign_state/2`: Updates the saga's state data.\n- `assign_context/2`: Updates the saga's context data.","ref":"SagaWeaver.Adapters.StorageAdapter.html#module-callbacks"},{"type":"behaviour","title":"Usage - SagaWeaver.Adapters.StorageAdapter","doc":"Developers can implement custom storage adapters by creating a module that implements this behaviour. The custom adapter must be configured in the application configuration.","ref":"SagaWeaver.Adapters.StorageAdapter.html#module-usage"},{"type":"behaviour","title":"Example Implementation - SagaWeaver.Adapters.StorageAdapter","doc":"```elixir\ndefmodule MyApp.CustomStorageAdapter do\n  @behaviour SagaWeaver.Adapters.StorageAdapter\n\n  alias SagaWeaver.SagaSchema\n\n  @impl true\n  def initialize_saga(saga_schema) do\n    # Custom logic to initialize and store the saga\n  end\n\n  @impl true\n  def saga_exists?(key) do\n    # Custom logic to check if the saga exists\n  end\n\n  # Implement other callbacks...\nend\n```","ref":"SagaWeaver.Adapters.StorageAdapter.html#module-example-implementation"},{"type":"behaviour","title":"Configuration - SagaWeaver.Adapters.StorageAdapter","doc":"To use a custom storage adapter, set it in your application's configuration:\n```elixir\nconfig :saga_weaver, :storage_adapter, MyApp.CustomStorageAdapter\n```\n\nThe impl/0 function in SagaWeaver.Adapters.StorageAdapter retrieves the configured adapter and delegates the function calls.","ref":"SagaWeaver.Adapters.StorageAdapter.html#module-configuration"},{"type":"behaviour","title":"Default Implementation - SagaWeaver.Adapters.StorageAdapter","doc":"By default, SagaWeaver may provide built-in adapters like RedisAdapter or PostgresAdapter. If none is configured, you need to specify one to enable saga persistence.","ref":"SagaWeaver.Adapters.StorageAdapter.html#module-default-implementation"},{"type":"function","title":"SagaWeaver.Adapters.StorageAdapter.assign_context/2","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#assign_context/2"},{"type":"callback","title":"SagaWeaver.Adapters.StorageAdapter.assign_context/2","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#c:assign_context/2"},{"type":"function","title":"SagaWeaver.Adapters.StorageAdapter.assign_state/2","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#assign_state/2"},{"type":"callback","title":"SagaWeaver.Adapters.StorageAdapter.assign_state/2","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#c:assign_state/2"},{"type":"function","title":"SagaWeaver.Adapters.StorageAdapter.complete_saga/1","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#complete_saga/1"},{"type":"callback","title":"SagaWeaver.Adapters.StorageAdapter.complete_saga/1","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#c:complete_saga/1"},{"type":"function","title":"SagaWeaver.Adapters.StorageAdapter.get_saga/1","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#get_saga/1"},{"type":"callback","title":"SagaWeaver.Adapters.StorageAdapter.get_saga/1","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#c:get_saga/1"},{"type":"function","title":"SagaWeaver.Adapters.StorageAdapter.initialize_saga/1","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#initialize_saga/1"},{"type":"callback","title":"SagaWeaver.Adapters.StorageAdapter.initialize_saga/1","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#c:initialize_saga/1"},{"type":"function","title":"SagaWeaver.Adapters.StorageAdapter.mark_as_completed/1","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#mark_as_completed/1"},{"type":"callback","title":"SagaWeaver.Adapters.StorageAdapter.mark_as_completed/1","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#c:mark_as_completed/1"},{"type":"function","title":"SagaWeaver.Adapters.StorageAdapter.saga_exists?/1","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#saga_exists?/1"},{"type":"callback","title":"SagaWeaver.Adapters.StorageAdapter.saga_exists?/1","doc":"","ref":"SagaWeaver.Adapters.StorageAdapter.html#c:saga_exists?/1"},{"type":"module","title":"SagaWeaver.Identifiers.DefaultIdentifier","doc":"The `SagaWeaver.Identifiers.DefaultIdentifier` module provides a default implementation for generating unique saga identifiers within the SagaWeaver framework.","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html"},{"type":"module","title":"Overview - SagaWeaver.Identifiers.DefaultIdentifier","doc":"This module implements the `SagaWeaver.Identifiers.SagaIdentifier` behaviour, supplying functions to:\n\n- Generate a unique saga ID based on a message, entity name, and a mapping.\n- Extract and map saga IDs from messages according to specified mappings.\n- Hash the extracted identifiers to produce a consistent and unique identifier.","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#module-overview"},{"type":"module","title":"Key Functions - SagaWeaver.Identifiers.DefaultIdentifier","doc":"- `unique_saga_id/3`: Generates a unique saga identifier.\n- `get_mapped_saga_ids/2`: Retrieves and maps saga IDs from a message.","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#module-key-functions"},{"type":"module","title":"Usage - SagaWeaver.Identifiers.DefaultIdentifier","doc":"This module is used internally by the SagaWeaver framework to ensure that each saga instance can be uniquely identified, preventing conflicts and ensuring messages are routed to the correct saga instance.","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#module-usage"},{"type":"module","title":"Examples - SagaWeaver.Identifiers.DefaultIdentifier","doc":"```elixir\nmessage = %MyApp.Events.OrderPlaced{order_id: 123}\nentity_name = :order_saga\nunique_saga_id_mapping = %{\n  MyApp.Events.OrderPlaced => fn msg -> %{order_id: msg.order_id} end\n}\n\nunique_id = SagaWeaver.Identifiers.DefaultIdentifier.unique_saga_id(\n  message,\n  entity_name,\n  unique_saga_id_mapping\n)\n# unique_id would be something like \"order_saga:abcdef123456...\"\n\n```","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#module-examples"},{"type":"function","title":"SagaWeaver.Identifiers.DefaultIdentifier.get_mapped_saga_ids/2","doc":"Retrieves and maps saga identifiers from a message based on the provided mapping.\n\nThis function looks up the message's struct in the `unique_saga_id_mapping` and applies the corresponding function to extract the identifiers.","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#get_mapped_saga_ids/2"},{"type":"function","title":"Parameters - SagaWeaver.Identifiers.DefaultIdentifier.get_mapped_saga_ids/2","doc":"- `message` (map): The message or event from which to extract identifiers.\n  - `unique_saga_id_mapping` (map): A mapping of message structs to functions that extract identifiers.","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#get_mapped_saga_ids/2-parameters"},{"type":"function","title":"Returns - SagaWeaver.Identifiers.DefaultIdentifier.get_mapped_saga_ids/2","doc":"- `saga_ids` (map): A map of identifiers extracted from the message.","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#get_mapped_saga_ids/2-returns"},{"type":"function","title":"Examples - SagaWeaver.Identifiers.DefaultIdentifier.get_mapped_saga_ids/2","doc":"iex> message = %MyApp.Events.OrderPlaced{order_id: 123}\n    iex> unique_saga_id_mapping = %{\n    ...>   MyApp.Events.OrderPlaced => fn msg -> %{order_id: msg.order_id} end\n    ...> }\n    iex> SagaWeaver.Identifiers.DefaultIdentifier.get_mapped_saga_ids(\n    ...>   message,\n    ...>   unique_saga_id_mapping\n    ...> )\n    %{order_id: 123}","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#get_mapped_saga_ids/2-examples"},{"type":"function","title":"SagaWeaver.Identifiers.DefaultIdentifier.unique_saga_id/3","doc":"Generates a unique saga identifier based on the message, entity name, and a mapping.\n\nThis function extracts identifiers from the message using the provided mapping, hashes them, and concatenates the hash with the entity name to produce a unique identifier.","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#unique_saga_id/3"},{"type":"function","title":"Parameters - SagaWeaver.Identifiers.DefaultIdentifier.unique_saga_id/3","doc":"- `message` (map): The message or event from which to extract identifiers.\n  - `entity_name` (atom): The name of the saga entity.\n  - `unique_saga_id_mapping` (map): A mapping of message structs to functions that extract identifiers.","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#unique_saga_id/3-parameters"},{"type":"function","title":"Returns - SagaWeaver.Identifiers.DefaultIdentifier.unique_saga_id/3","doc":"- `unique_id` (String.t): A unique identifier for the saga instance.","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#unique_saga_id/3-returns"},{"type":"function","title":"Examples - SagaWeaver.Identifiers.DefaultIdentifier.unique_saga_id/3","doc":"iex> message = %MyApp.Events.OrderPlaced{order_id: 123}\n    iex> entity_name = :order_saga\n    iex> unique_saga_id_mapping = %{\n    ...>   MyApp.Events.OrderPlaced => fn msg -> %{order_id: msg.order_id} end\n    ...> }\n    iex> SagaWeaver.Identifiers.DefaultIdentifier.unique_saga_id(\n    ...>   message,\n    ...>   entity_name,\n    ...>   unique_saga_id_mapping\n    ...> )\n    \"order_saga:a1b2c3d4e5f6g7h8i9j0...\"","ref":"SagaWeaver.Identifiers.DefaultIdentifier.html#unique_saga_id/3-examples"},{"type":"behaviour","title":"SagaWeaver.Identifiers.SagaIdentifier","doc":"This module is responsible for generating unique identifiers for sagas.","ref":"SagaWeaver.Identifiers.SagaIdentifier.html"},{"type":"function","title":"SagaWeaver.Identifiers.SagaIdentifier.get_mapped_saga_ids/2","doc":"","ref":"SagaWeaver.Identifiers.SagaIdentifier.html#get_mapped_saga_ids/2"},{"type":"callback","title":"SagaWeaver.Identifiers.SagaIdentifier.get_mapped_saga_ids/2","doc":"","ref":"SagaWeaver.Identifiers.SagaIdentifier.html#c:get_mapped_saga_ids/2"},{"type":"function","title":"SagaWeaver.Identifiers.SagaIdentifier.unique_saga_id/3","doc":"","ref":"SagaWeaver.Identifiers.SagaIdentifier.html#unique_saga_id/3"},{"type":"callback","title":"SagaWeaver.Identifiers.SagaIdentifier.unique_saga_id/3","doc":"","ref":"SagaWeaver.Identifiers.SagaIdentifier.html#c:unique_saga_id/3"},{"type":"module","title":"SagaWeaver.Orchestrator","doc":"The `SagaWeaver.Orchestrator` module is responsible for orchestrating the execution of sagas within the SagaWeaver framework. It handles the initiation, retrieval, execution, and completion of saga instances based on incoming messages.","ref":"SagaWeaver.Orchestrator.html"},{"type":"module","title":"Overview - SagaWeaver.Orchestrator","doc":"- **Execute Saga**: Determines whether to start a new saga or continue an existing one based on the incoming message.\n- **Handle Saga**: Processes the saga by invoking the appropriate handler and manages its state.\n- **Start Saga**: Initializes a new saga instance if the message is eligible to start a saga.\n- **Initialize Saga**: Sets up the initial state of a new saga and stores it using the configured storage adapter.\n- **Retrieve Saga**: Fetches an existing saga instance based on a unique identifier derived from the message.","ref":"SagaWeaver.Orchestrator.html#module-overview"},{"type":"module","title":"Key Functions - SagaWeaver.Orchestrator","doc":"- `execute_saga/2`: Entry point for processing messages and orchestrating sagas.\n- `handle_saga/3`: Handles the saga logic, updating its state, and marking it as completed if necessary.\n- `start_saga/2`: Checks if a message can start a new saga and initializes it.\n- `initialize_saga/2`: Creates and stores a new saga instance.\n- `retrieve_saga/2`: Retrieves an existing saga instance based on the message.","ref":"SagaWeaver.Orchestrator.html#module-key-functions"},{"type":"module","title":"Usage - SagaWeaver.Orchestrator","doc":"The `Orchestrator` module is typically not used directly by end-users but is invoked by the SagaWeaver framework when messages are processed.","ref":"SagaWeaver.Orchestrator.html#module-usage"},{"type":"module","title":"Examples - SagaWeaver.Orchestrator","doc":"```elixir\n# Assuming `MySaga` is a module that uses SagaWeaver\nmessage = %MyApp.SomeEvent{}\nSagaWeaver.Orchestrator.execute_saga(MySaga, message)\n```","ref":"SagaWeaver.Orchestrator.html#module-examples"},{"type":"module","title":"Dependencies - SagaWeaver.Orchestrator","doc":"SagaWeaver.Adapters.StorageAdapter: Interface for storage operations.\nSagaWeaver.Identifiers.SagaIdentifier: Generates unique identifiers for sagas.\nSagaWeaver.SagaSchema: Defines the saga's data structure.","ref":"SagaWeaver.Orchestrator.html#module-dependencies"},{"type":"function","title":"SagaWeaver.Orchestrator.execute_saga/2","doc":"Executes a saga by processing the given message.\n\nThis function serves as the entry point for handling messages within a saga context. It determines whether to start a new saga or continue an existing one based on the message.","ref":"SagaWeaver.Orchestrator.html#execute_saga/2"},{"type":"function","title":"Parameters - SagaWeaver.Orchestrator.execute_saga/2","doc":"- `saga_module` (atom): The saga module that defines the saga logic.\n  - `message` (map): The message or event to be processed.","ref":"SagaWeaver.Orchestrator.html#execute_saga/2-parameters"},{"type":"function","title":"Returns - SagaWeaver.Orchestrator.execute_saga/2","doc":"- `{:ok, saga_instance}`: Indicates the saga was successfully processed.\n  - `{:noop, reason}`: No operation was performed, with a reason provided.","ref":"SagaWeaver.Orchestrator.html#execute_saga/2-returns"},{"type":"function","title":"Examples - SagaWeaver.Orchestrator.execute_saga/2","doc":"iex> SagaWeaver.Orchestrator.execute_saga(MyApp.OrderSaga, message)\n    {:ok, %SagaSchema{}}","ref":"SagaWeaver.Orchestrator.html#execute_saga/2-examples"},{"type":"function","title":"SagaWeaver.Orchestrator.initialize_saga/2","doc":"Initializes a new saga instance with initial state.\n\nCreates a new saga instance using the provided message and stores it using the configured storage adapter.","ref":"SagaWeaver.Orchestrator.html#initialize_saga/2"},{"type":"function","title":"Parameters - SagaWeaver.Orchestrator.initialize_saga/2","doc":"- `saga_module` (atom): The saga module.\n  - `message` (map): The message that starts the saga.","ref":"SagaWeaver.Orchestrator.html#initialize_saga/2-parameters"},{"type":"function","title":"Returns - SagaWeaver.Orchestrator.initialize_saga/2","doc":"- `{:ok, saga_instance}`: The new saga instance.\n  - `{:ok, :not_found}`: No saga instance was found.","ref":"SagaWeaver.Orchestrator.html#initialize_saga/2-returns"},{"type":"function","title":"Examples - SagaWeaver.Orchestrator.initialize_saga/2","doc":"iex> SagaWeaver.Orchestrator.initialize_saga(MyApp.OrderSaga, start_message)\n    {:ok, %SagaSchema{}}","ref":"SagaWeaver.Orchestrator.html#initialize_saga/2-examples"},{"type":"function","title":"SagaWeaver.Orchestrator.retrieve_saga/2","doc":"Retrieves an existing saga instance based on the message.\n\nGenerates a unique saga identifier from the message and attempts to retrieve the saga from storage.","ref":"SagaWeaver.Orchestrator.html#retrieve_saga/2"},{"type":"function","title":"Parameters - SagaWeaver.Orchestrator.retrieve_saga/2","doc":"- `saga_module` (atom): The saga module.\n  - `message` (map): The message associated with the saga.","ref":"SagaWeaver.Orchestrator.html#retrieve_saga/2-parameters"},{"type":"function","title":"Returns - SagaWeaver.Orchestrator.retrieve_saga/2","doc":"- `{:ok, saga_instance}`: The existing saga instance.\n  - `{:ok, :not_found}`: No saga instance was found.","ref":"SagaWeaver.Orchestrator.html#retrieve_saga/2-returns"},{"type":"function","title":"Examples - SagaWeaver.Orchestrator.retrieve_saga/2","doc":"iex> SagaWeaver.Orchestrator.retrieve_saga(MyApp.OrderSaga, message)\n    {:ok, %SagaSchema{}}","ref":"SagaWeaver.Orchestrator.html#retrieve_saga/2-examples"},{"type":"function","title":"SagaWeaver.Orchestrator.start_saga/2","doc":"Starts a new saga if the message is eligible to initiate one.\n\nThis function checks whether the incoming message is among those that can start a new saga, as defined by the `started_by/0` callback in the saga module.","ref":"SagaWeaver.Orchestrator.html#start_saga/2"},{"type":"function","title":"Parameters - SagaWeaver.Orchestrator.start_saga/2","doc":"- `saga_module` (atom): The saga module.\n  - `message` (map): The message that may start a new saga.","ref":"SagaWeaver.Orchestrator.html#start_saga/2-parameters"},{"type":"function","title":"Returns - SagaWeaver.Orchestrator.start_saga/2","doc":"- `{:ok, saga_instance}`: A new saga has been initialized.\n  - `{:noop, reason}`: The message does not start a saga.","ref":"SagaWeaver.Orchestrator.html#start_saga/2-returns"},{"type":"function","title":"Examples - SagaWeaver.Orchestrator.start_saga/2","doc":"iex> SagaWeaver.Orchestrator.start_saga(MyApp.OrderSaga, start_message)\n    {:ok, %SagaSchema{}}","ref":"SagaWeaver.Orchestrator.html#start_saga/2-examples"},{"type":"module","title":"SagaWeaver.Saga","doc":"The `SagaWeaver.Saga` module provides a macro to simplify the creation of sagas within the SagaWeaver framework. By using `SagaWeaver.Saga`, developers can define sagas that orchestrate complex, long-running transactions across multiple services or operations.","ref":"SagaWeaver.Saga.html"},{"type":"module","title":"Overview - SagaWeaver.Saga","doc":"A saga represents a sequence of operations that must be executed reliably and consistently, often involving compensating actions in case of failures. This module sets up the necessary boilerplate, allowing you to focus on implementing the specific logic of your saga.\n\nWhen you `use SagaWeaver.Saga`, the following happens:\n\n- The module is set to implement the `SagaWeaver.SagaBehaviour` behaviour.\n- Default implementations for required callbacks are provided.\n- Helper functions are injected to manage saga state and context.","ref":"SagaWeaver.Saga.html#module-overview"},{"type":"module","title":"Usage - SagaWeaver.Saga","doc":"To define a saga, create a module and `use SagaWeaver.Saga`, optionally providing configuration options:\n\n```elixir\ndefmodule MyApp.OrderSaga do\n  use SagaWeaver.Saga,\n    started_by: [MyApp.Events.OrderPlaced],\n    identity_key_mapping: %{\n      MyApp.Events.OrderPlaced => fn message -> %{order_id: message.order_id} end\n    }\n\n  @impl true\n  def handle_message(saga_instance, %MyApp.Events.OrderPlaced{} = message) do\n    # Implement your saga logic here\n    saga_instance = assign_state(saga_instance, :order_status, :placed)\n    {:ok, saga_instance}\n  end\n\n  # Optionally override other callbacks or functions as needed\nend\n\n```","ref":"SagaWeaver.Saga.html#module-usage"},{"type":"module","title":"Configuration Options - SagaWeaver.Saga","doc":"- :started_by (optional): A list of message modules that can initiate this saga. Defaults to an empty list.\n- :identity_key_mapping (optional): A mapping that defines how to extract identifiers from messages to uniquely identify saga instances. Defaults to an empty map.","ref":"SagaWeaver.Saga.html#module-configuration-options"},{"type":"module","title":"Callbacks Implemented - SagaWeaver.Saga","doc":"The following callbacks from SagaWeaver.SagaBehaviour are implemented with default or configurable behavior:\n\n- entity_name/0: Returns the module name as the saga's entity name.\n- started_by/0: Returns the list of message modules that can start this saga.\n- identity_key_mapping/0: Returns the mapping used to identify the saga instance.\n- handle_message/2: Handles incoming messages and updates the saga's state. Defaults to returning an error if not overridden.","ref":"SagaWeaver.Saga.html#module-callbacks-implemented"},{"type":"module","title":"Helper Functions - SagaWeaver.Saga","doc":"The module provides several functions to manage saga state and context:\n\n- assign_state/2: Updates the saga's state with a map of key-value pairs.\n- assign_state/3: Updates the saga's state by setting a single key-value pair.\n- assign_context/2: Updates the saga's context with a map of key-value pairs.\n- assign_context/3: Updates the saga's context by setting a single key-value pair.\n- mark_as_completed/1: Marks the saga as completed.\nThese functions interact with the underlying storage adapter to persist changes.","ref":"SagaWeaver.Saga.html#module-helper-functions"},{"type":"module","title":"Overridable Functions - SagaWeaver.Saga","doc":"The following functions are marked as overridable, allowing you to provide custom implementations:\n\n- handle_message/2\n- started_by/0\n- entity_name/0\n- identity_key_mapping/0","ref":"SagaWeaver.Saga.html#module-overridable-functions"},{"type":"extras","title":"README","doc":"[![Coverage Status](https://coveralls.io/repos/github/Hermanlangner/saga_weaver/badge.svg?branch=main)](https://coveralls.io/github/Hermanlangner/saga_weaver?branch=main)\n# SagaWeaver\n\nA library to help you execute distributed transactions using Sagas without needing to worry about transport layers or storage implementations.","ref":"readme.html"},{"type":"extras","title":"Inspiration - README","doc":"There are many situations where a saga can be created from 1 or more events, either through an external system or internal. Usually the pattern is implemented as part of a framework or messaging library. The goal of saga weaver is to let you hook into any method of transport (message queues or http) and still be able to run a saga.\nStorage can be a big source of race conditions, that bring a lot of mental overhead while rolling your own. Saga weaver is made to take care of that layer for you by providing storage implementations or if that's not good enough, expose an interface for you to write your own if you would like a more efficient approach or use a different pattern.\n\nDistributed transactions are already hard on their own, Saga Weaver lets you focus on building out your application rules rather than spending time on covering storage or transport race conditions.","ref":"readme.html#inspiration"},{"type":"extras","title":"Design Approach - README","doc":"- All Saga Weaver adapters are built with optimistic concurrency as a first class citizen. While collisions are guarenteed, the most frequent scenario happens in a fan out fan in. Otherwise generally collisions are low.\n- Life Cycles are fully configurable, The conditions to start a saga, fork a workflow, trigger a compensating transaction or close a saga are all simply your elixir code.\n- Setting state and context is intended to `feel` like the live view flow, so that there's familiarity with the rest of your eco system.\n- Any struct that participates in a Saga needs to be able to be transformed to the Saga identifier\n- While an Inbox/Outbox pattern is not present, we follow that at least once delivery approach. It needs to be ensured that application logic supports it.","ref":"readme.html#design-approach"},{"type":"extras","title":"Installation - README","doc":"SagaWeaver is published on [Hex](https://hexdocs.pm/saga_weaver), the package can be installed\nby adding `saga_weaver` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:saga_weaver, \"~> 0.2\"}\n  ]\nend\n```\n\n- Add SagaWeaver to your `application.ex`\n\n```elixir\nchildren = [\n      {SagaWeaver, []}\n    ]\n ```","ref":"readme.html#installation"},{"type":"extras","title":"Postgress - README","doc":"- If you're using Postgres, you need to generate a migration to setup the saga_weaver table\n\n```elixir\n  mix ecto.gen.migration add_saga_weaver_table\n```\n\nThen add the following to your migration\n\n```elixir\n  def change do\n    create table(:sagaweaver_sagas) do\n      add(:uuid, :string)\n      add(:saga_name, :string)\n      add(:states, :map, default: %{})\n      add(:context, :map, default: %{})\n      add(:marked_as_completed, :boolean, default: false)\n      add(:lock_version, :integer, default: 1)\n\n      timestamps()\n    end\n\n    create(index(:sagaweaver_sagas, [:uuid], unique: true))\n  end\n```\n\n- Add the config to `config.exs` and set your Repo and Adapter\n\n```elixir\nconfig :saga_weaver, SagaWeaver,\n  storage_adapter: SagaWeaver.Adapters.PostgresAdapter,\n  repo: MyApp.Repo\n```","ref":"readme.html#postgress"},{"type":"extras","title":"Redis - README","doc":"add your redis config to `config.exs`\n\n```elixir\nconfig :saga_weaver, SagaWeaver,\n  host: \"localhost\",\n  port: 6379,\n  namespace: \"my_app\",\n  storage_adapter: SagaWeaver.Adapters.RedisAdapter,\n```","ref":"readme.html#redis"},{"type":"extras","title":"Example - README","doc":"A simple version of a saga, with one created message and one close message. We can define it as follows\n\n```elixir\ndefmodule StartSagaMessage do\n  defstruct [:id, :name]\nend\n\ndefmodule CloseSagaMessage do\n  defstruct [:external_id, :fanout_id]\nend\n\ndefmodule SimpleSaga do\n  use SagaWeaver.Saga,\n    started_by: [StartSagaMessage],\n    identity_key_mapping: %{\n      StartSagaMessage => fn message -> %{id: message.id} end,\n      CloseSagaMessage => fn message -> %{id: message.external_id} end\n    }\n\n  alias SagaWeaver.SagaSchema\n\n  def handle_message(%SagaSchema{} = instance, %StartSagaMessage{} = message) do\n    case instance.states[\"start_handled\"] do\n      true ->\n        IO.puts(\"Start Message already handled for id: #{message.id}\")\n\n      _other ->\n        IO.puts(\"Starting Saga for id: #{message.id}\")\n        # Do initial setup\n    end\n\n    {:ok,\n     instance\n     |> assign_state(\"start_handled\", true)}\n  end\n\n  def handle_message(%SagaSchema{} = instance, %CloseSagaMessage{} = message) do\n    instance = instance |> assign_state(\"close_handled\", true)\n\n    if ready_to_complete?(instance) do\n      IO.puts(\"All conditions for closure have been met, closing\")\n      {:ok, instance |> mark_as_completed()}\n    else\n      {:ok, instance}\n    end\n  end\n\n  defp ready_to_complete?(instance) do\n    instance.states[\"start_handled\"] && instance.states[\"close_handled\"]\n  end\nend\n\n```\n\nThis basic saga gets started with the Start message and closes when the accompanying Close message happens.\nLet's see how this pans out.\n\n- Ensure you're fully migrated `mix ecto.migrate`\n- Open up your shell `iex -S mix`\n- Setup some structs from the above example\n\n```elixir\nstart_message = %StartSagaMessage{id: 1, name: \"started\"}\nclose_message = %CloseSagaMessage{external_id: 1, fanout_id: 24}\nfake_close_message = %CloseSagaMessage{external_id: 2, fanout_id: 23}\n```\n\nLet's try and start a saga.\nRun\n\n```elixir\nSagaWeaver.execute_saga(SimpleSaga, start_message)\n```\n\nThe should be output for\n`Starting Saga for id: 1`\n\nIf you run the command a second time, you should get\n`Start Message already handled for id: 1`\n\nLets try and handle a close message that can't be associated to the saga\n\n```elixir\nSagaWeaver.execute_saga(SimpleSaga, fake_close_message)\n```\n\nyou will get\n\n```elixir\n{:noop,\n \"No active Sagas were found for this message, this message also does not start a new Saga.\"}\n```\n\nLet's trigger a message that will close the Saga\n\n```elixir\nSagaWeaver.execute_saga(SimpleSaga, close_message)\n```\n\n`All conditions for closure have been met, closing`\n\n## `use Saga`\n\nA breakdown on how to use a saga. All you need is a starting message, and an identity map\n\n```elixir\n    use SagaWeaver.Saga,\n      started_by: [StartSagaMessage],\n      identity_key_mapping: %{\n        StartSagaMessage => fn message -> %{id: message.id} end,\n        CloseSagaMessage => fn message -> %{id: message.external_id} end\n      }\n\n```","ref":"readme.html#example"},{"type":"extras","title":"Started By - README","doc":"`started_by` is the bread and butter of SagaWeaver. To have a valid Saga you need at least 1 struct that starts a Saga. It's possible for all messages to start a saga, and there is no required order. You can handle it through setting the state to co-ordinate the transaction.\n\n```elixir\nready_for_next_step = instance.states[\"start_message_1\"] && instance.states[\"start_message_2\"]\n\nif read_for_next_step do","ref":"readme.html#started-by"},{"type":"extras","title":"Logic to kick off next step process - README","doc":"end\n```","ref":"readme.html#logic-to-kick-off-next-step-process"},{"type":"extras","title":"identity_key_mapping - README","doc":"In order to function with the default SagaWeaver configuration, each struct that is handled needs a map setup to extract elements to uniquely identify the saga\nIn the format\n\n```elixir\nmy_identity_mapping = %{\n  MyMessageModule => function_to_extract_context.(message)\n}\n```\n\nIt is required for scenarios where external Api's or domains have a different name for the identifier and you need the ability to correlate them.\nIn the future more sensible defaults could be worth it.","ref":"readme.html#identity_key_mapping"},{"type":"extras","title":"Setting State - README","doc":"Each handler is passed an instance of a Saga where you can set states to manage your lifecycle.\nSee more here - add link to hex.pm","ref":"readme.html#setting-state"},{"type":"extras","title":"Setting Context - README","doc":"Context is for information that's not related to managing the state of the saga, but is needed to trigger events or possibly combined on closure\nSee more here - add link to hex.pm","ref":"readme.html#setting-context"},{"type":"extras","title":"Completing A saga - README","doc":"When either all your conditions for completion or rolling back has been completed. The final action you need to do is complete your saga. It's the equivalent to commiting a transaction or completing a rollback in sql.\n\n```elixir\n instance |> mark_as_completed()\n```\n\nIf sagas are not closed, the transaction will never commit and until timeouts are added will have your application in a \"stuck\" state.","ref":"readme.html#completing-a-saga"},{"type":"extras","title":"Roadmap - README","doc":"**Developer Experience & Quality**\n\n- [ ] Clean up tests to run for multiple adapters as a suite: Enhance the test suite to ensure compatibility and reliability across different storage adapters.\n- [ ] Generate migration for setup: Provide mix tasks to generate necessary database migrations for easy setup.\n- [ ] Make atoms work correctly for PostgreSQL adapters: Ensure that atoms are handled appropriately when using PostgreSQL as the storage backend.\n- [ ] Ensure workflows are as expected if saga is killed while processing: Improve fault tolerance by handling unexpected terminations gracefully.\n- [ ] Adopt Nimble Config for better configs: Utilize NimbleConfig for more robust and flexible configuration management.\n\n**Features**\n\n- [ ] Allow capability to schedule timeouts for timeouts: Implement functionality to handle timeouts, allowing sagas to be scheduled for timeout actions.\n- [ ] Store historic completed sagas with a TTL: Enable storage of completed sagas with a Time-To-Live (TTL) to retain history for a configurable duration.\n- [ ] Add observability queries: Introduce built-in queries to monitor saga executions and states for better observability.\n- [ ] Add alerting API: Provide an API for setting up alerts based on saga events or failures.\n- [ ] Add telemetry: Integrate with Telemetry to offer insights into saga performance and metrics.\n- [ ] Add UI to monitor and track sagas: Develop a web-based dashboard to visualize and manage sagas in real-time.\n- [ ] Add SQLite adapter: Expand storage options by adding support for SQLite.\n- [ ] Add native Elixir adapter: Implement an in-memory adapter for testing or lightweight use cases without external dependencies.\n- [ ] Allow configurable retries and timeouts\n- [ ] Add pooling to Redis Adapter","ref":"readme.html#roadmap"},{"type":"extras","title":"CHANGELOG","doc":"","ref":"changelog.html"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}